local utils = require "LCV.utils"

local function checkEmpty(img)
	for y, row in ipairs(img) do
		for x, val in ipairs(row) do
			if val == 1 then
				return false
			end
		end
	end
	return true
end

local function findParent(label, list)
	parent = label
	while(list[parent] ~= 0 and list[parent] ~= parent) do
		parent = list[parent]
	end
	return parent
end

local function labelComponents(input)
	if checkEmpty(input) then 
		return utils.deepcopy(input), 0
	end
	local h = #input
	local w = #input[1]
	local listParents = {}
	local labelCounter = 1

	local output = {}
	for y = 1, h do
		output[y] = {}
		for x = 1, w do
			local neighbors = {}
			if input[y][x] >= 1 then
				if x ~= 1 and input[y][x-1] >= 1 then
					table.insert(neighbors, {x = x-1, y = y})
				end

				if x ~= 1 and y ~= 1 and input[y -1][x-1] >= 1 then
					table.insert(neighbors, {x = x-1, y = y-1})
				end

				if y ~= 1 and input[y -1][x] >= 1 then
					table.insert(neighbors, {x = x, y = y-1})
				end

				if x ~= w and y ~= 1 and input[y -1][x+1] >= 1 then
					table.insert(neighbors, {x = x+1, y = y-1})
				end

				if #neighbors > 0 then 
					output[y][x] = output[neighbors[1].y][neighbors[1].x]
				else 
					output[y][x] = labelCounter
					listParents[output[y][x]] = 0
					labelCounter = labelCounter + 1
				end

				if #neighbors > 1 then
					for i = 1, #neighbors do
						local label = output[neighbors[i].y][neighbors[i].x]
						if not (label == output[y][x]) then
							if not (listParents[label] == output[y][x] or listParents[output[y][x]] == label) then
								if listParents[label] ~= 0 then
									if listParents[output[y][x]] ~= 0 then
										listParents[findParent(label, listParents)] = output[y][x]
									else 
										listParents[output[y][x]] = label
									end
								else
									listParents[label] = output[y][x]
								end
							end
						end
					end
				end
			else
				output[y][x] = 0
			end
		end
	end

	for y = 1, h do 
		for x = 1, w do
			output[y][x] = findParent(output[y][x], listParents) or 0
		end
	end
	local map = {}
	local counter = 1
	for y, row in ipairs(output) do
		for x, label in ipairs(row) do
			if(label == 0) then
				output[y][x] = 0
			else
				if map[label] then
					output[y][x] = map[label]
				else
					output[y][x] = counter
					map[label] = counter
					counter = counter + 1
				end
			end 
		end
	end
	return output, counter - 1
end

local function dilate(imgA, imgB)
	local imgOut = {}
	for y, row in ipairs(imgA) do
		imgOut[y] = {}
		for x, val in ipairs(row) do
			imgOut[y][x] = 0
		end
	end

	for y, row in ipairs(imgA) do
		for x, val in ipairs(row) do
			if val == 1 then
				local h = #imgB
				local w = #imgB[1]

				local centerX = math.ceil(w/2)
				local centerY = math.ceil(h/2)
				for yB, rowB in ipairs(imgB) do
					for xB, valB in ipairs(rowB) do
						local testX = x - centerX + xB
						local testY = y - centerY + yB
						if (imgOut[testY]) then
							if (imgOut[testY][testX] == 0) then
								imgOut[testY][testX] = imgB[yB][xB]
							end
						end
					end
				end
			end
		end
	end
	return imgOut
end

local function erode(imgA, imgB)
	local imgOut = {}
	for y, row in ipairs(imgA) do
		imgOut[y] = {}
		for x, val in ipairs(row) do
			imgOut[y][x] = 0
		end
	end

	for y, row in ipairs(imgA) do
		for x, val in ipairs(row) do
			if val == 1 then
				imgOut[y][x] = 1
				local h = #imgB
				local w = #imgB[1]

				local centerX = math.ceil(w/2)
				local centerY = math.ceil(h/2)
				for yB, rowB in ipairs(imgB) do
					for xB, valB in ipairs(rowB) do
						local testX = x - centerX + xB
						local testY = y - centerY + yB
						if (imgOut[testY]) and (imgOut[testY][testX]) then
							--print(testX, testY, imgOut[testY][testX])
							if not (imgA[testY][testX] >= imgB[yB][xB]) then
								imgOut[y][x] = 0
							end
						elseif (imgB[yB][xB] == 1) then
							imgOut[y][x] = 0
						end
					end
				end
			end
		end
	end
	return imgOut
end

local function open(imgA, imgB)
	return dilate(erode(imgA, imgB), imgB)
end

local function close(imgA, imgB)
	return erode(dilate(imgA, imgB), imgB)
end

--assumes same size
local function difference(imgA, imgB)
	imgOut = {}
	for y, row in ipairs(imgA) do
		imgOut[y] = {}
		for x, val in ipairs(row) do
			if (imgA[y][x] == imgB[y][x]) then
				imgOut[y][x] = 0
			else
				imgOut[y][x] = imgA[y][x] 
			end
		end
	end 
	return imgOut
end

local function toBinary(imgData)
	local h = imgData:getHeight()
	local w = imgData:getWidth()
	local img = {}
	for y=1, h do
		img[y] = {}
		for x=1, w do
			local r,g,b = imgData:getPixel(x-1,y-1)
			if r < .5 and g <.5 and b  <.5 then
				img[y][x] = 1
			else
				img[y][x] = 0
			end
		end
	end
	return img
end

local function generateB(n)
	local unit = {}
	for y=1, 65 do
		unit[y] = {}
		for x=1, 65 do
			if n > 0 then
				if (x == 33 and y == 33) or (x == 32 and y == 33) or (x == 34 and y == 33) 
					or (x == 33 and y == 32) or (x == 33 and y == 34) then			
					unit[y][x] = 1				
				else
					unit[y][x] = 0
				end
			else
				if(x == 33 and y == 33) then
					unit[y][x] = 1
				end
			end
		end
	end

	local b = utils.deepcopy(unit)
	local i = 1
	while (i < n ) do
		b = dilate(b,unit)
		i = i + 1
	end
	return b
end

local function generateSkeletalSet(img, b)
	local unit = {}
	for y=1, 65 do
		unit[y] = {}
		for x=1, 65 do

			if (x == 33 and y == 33) or (x == 32 and y == 33) or (x == 34 and y == 33) 
				or (x == 33 and y == 32) or (x == 33 and y == 34) then			
				unit[y][x] = 1				
			else
				unit[y][x] = 0
			end
		end
	end

	local part1 = erode(img, b)
	local part2 = open(part1, unit)

	return difference(part1, part2)
end

local function union(imgA, imgB)
	imgOut = {}
	for y, row in ipairs(imgA) do
		imgOut[y] = {}
		for x, val in ipairs(row) do
			if (imgA[y][x] == 1 or imgB[y][x] == 1) then
				imgOut[y][x] = 1
			else
				imgOut[y][x] = 0 
			end
		end
	end 
	return imgOut
end

--takes binary image or grayscale.
local function getRawMoment(img, i, j)
	local m = 0
	for y, row in ipairs(img) do
		for x, val in ipairs(row) do
			m = m + (x^i * y^j) * img[y][x]
		end
	end
	return m 
end

local function getArea(img)
	local area = getRawMoment(img, 0, 0)
	return area
end

local function getCentriod(img)
	local area = getRawMoment(img, 0, 0)
	local xcenter = getRawMoment(img, 1, 0) / area 
	local ycenter = getRawMoment(img, 0, 1) / area

	return xcenter, ycenter
end

local function getCentralMoment(img, i, j)

	local area = getRawMoment(img, 0, 0)
	local xcenter = getRawMoment(img, 1, 0) / area 
	local ycenter = getRawMoment(img, 0, 1) / area

	local m = 0
	for y, row in ipairs(img) do
		for x, val in ipairs(row) do
			m = m + ((x - xcenter)^i * (y - ycenter)^j) * img[y][x]
		end
	end
	return m 

end

-- takes an image returns a list of images of the different componenets.
local function getComponents(img)
	local labeledImage, numComponents = labelComponents(img)

	local images = {}
	for i=1, numComponents do
		images[i] = {}
	end

	for y, row in ipairs(labeledImage) do
		for i=1, numComponents do
			images[i][y] = {}
		end
		for x, val in ipairs(row) do
			for i=1, numComponents do
				if val == i then
					images[i][y][x] = 1
				else
					images[i][y][x] = 0
				end
			end
		end
	end
	return images
end

return {
	labelComponents = labelComponents,
	dilate = dilate,
	erode = erode,
	open = open,
	close = close,
	difference = difference,
	toBinary = toBinary,
	generateB = generateB,
	generateSkeletalSet = generateSkeletalSet,
	union = union,
	checkEmpty = checkEmpty,
	getRawMoment = getRawMoment,
	getCentralMoment = getCentralMoment,
	getCentriod = getCentriod,
	getArea = getArea,
	getComponents = getComponents,

}