local LCV = require "LCV.LCV"
local utils = require "LCV.utils"

local function getDirection(xCenter,yCenter,  x,y)
	local angle = math.atan2 ( yCenter - y, x - xCenter )
	return angle
end

--takes binary images
local function findNumber(img)
	local largeStructure = LCV.generateB(14)
	local smallStructure = LCV.generateB(2)
	local baysImage = LCV.difference(LCV.close(img, largeStructure), img)
	local bays = LCV.getComponents(baysImage)

	local dataTable = {
		bayDirections = {

		},
		bayLocations = {

		},
		lakeLocations = {

		}
	}


	for i,bay in ipairs(bays) do
		local xCenter, yCenter = LCV.getCentriod(bay)
		lip = LCV.difference(LCV.difference(LCV.dilate(bay, smallStructure), img), bay)
		local _, count = LCV.labelComponents(lip)
		if count == 0 then 
			table.insert(dataTable.lakeLocations, {xCenter, yCenter})
		else
			if LCV.getArea(bay) > 1 then
				table.insert(dataTable.bayLocations, {xCenter, yCenter})
				local x,y = LCV.getCentriod(lip)
				table.insert(dataTable.bayDirections, getDirection(xCenter, yCenter, x, y))
			end
		end
	end

	--the easy stuff
	if #dataTable.bayLocations == 0 and #dataTable.lakeLocations == 1 then 
		return 0
	elseif #dataTable.bayLocations == 0 and #dataTable.lakeLocations == 0 then 
		return 1
	elseif #dataTable.bayLocations == 3 and #dataTable.lakeLocations == 0 then 	
		return 1
	elseif #dataTable.bayLocations == 2 and #dataTable.lakeLocations == 1 then 
		return 2
	elseif #dataTable.bayLocations == 2 and #dataTable.lakeLocations == 2 then 
		return 8
	elseif #dataTable.bayLocations == 1 and #dataTable.lakeLocations == 0 then 
		return 7
	end

	local imgX, imgY = LCV.getCentriod(img)
	if #dataTable.bayLocations == 2 then 
		if dataTable.bayLocations[1][1] < imgX and dataTable.bayLocations[2][1] < imgX then 
			local angleRatio = math.abs(dataTable.bayDirections[1]/dataTable.bayDirections[2])
			if angleRatio < 1.2 and angleRatio > .8 then
				return 3
			else
				return 4
			end
		else 
			return 5
		end
	elseif #dataTable.bayLocations == 1 and #dataTable.lakeLocations == 1 then 
		if dataTable.bayDirections[1] > 0 then 
			return 6
		else
			return 9
		end
	elseif #dataTable.bayLocations == 3 and #dataTable.lakeLocations == 1 then 
		if dataTable.lakeLocations[1][2] < imgY then 
			return 4 
		else
			return 2	    
		end 
	end
	print(utils.serial(dataTable))

end

function  love.load( ... )
	local image = LCV.toBinary(love.image.newImageData( "unknown.png" ))
	local num = findNumber(image)
	print(num)
end


